#include "index.h"

const char *index_mxfields[2] = {"dd"};

Index index_new(int nn) {
  Index oo;
  oo.nn = nn;
  oo.dd  = mxMalloc(sizeof(int) * nn);
  return oo;
}

void index_free(Index ii) {
  mxFree(ii.dd);
}

Index index_copy(Index ii) {
  Index oo;
  int kk;
  oo = index_new(ii.nn);
  for ( kk = 0 ; kk < oo.nn ; kk ++ ) {
    oo.dd[kk]  = ii.dd[kk];
  }
  return oo;
}


Index index_readmxarray(const mxArray *marray) {
  Index oo;
  double *dp, dd;
  int kk, nn, ii; 
  if ( mxIsClass(marray,"index") || mxIsClass(marray,"shape") || 
       mxIsClass(marray,"table") ) {
    oo = index_new(mxGetNumberOfElements(mxGetField(marray,0,"dd")));
    dp = mxGetPr(mxGetField(marray,0,"dd"));
    for ( kk = 0 ; kk < oo.nn ; kk ++ ) {
      oo.dd[kk]  = * ( dp ++ );
    }
  } else { /* double list */
    nn = mxGetNumberOfElements(marray);
    oo = index_new(nn);
    dp = mxGetPr(marray);
    for ( kk = 0 ; kk < nn ; kk ++ ) {
      oo.dd[kk]  = * ( dp ++ );
      for ( ii = 0 ; ii < kk ; ii ++ ) { /* need to make sure dd sorted */
        if ( oo.dd[ii] > oo.dd[kk] ) {
          dd        = oo.dd[ii];
          oo.dd[ii] = oo.dd[kk];
          oo.dd[kk] = dd;
        }
      }
    }
  }
  return oo;
}

mxArray *index_writemxarray(Index ii) {
  mxArray *marray, *de;
  double *dep;
  int kk;

  marray = mxCreateStructMatrix(1,1,1,index_mxfields);
  de  = mxCreateDoubleMatrix(1,ii.nn,mxREAL);
  dep = mxGetPr(de);
  for ( kk = 0 ; kk < ii.nn ; kk ++ ) {
    * ( dep ++ ) = ii.dd[kk];
  }
  mxSetField(marray,0,"dd", de);
  mexCallMATLAB(1,&de,1,&marray,"index");
  mxDestroyArray(marray);
  return de;
}

void index_rewritemxarray(mxArray *marray, Index ii) {
  mxArray *de;
  double *dep;
  int kk;

  if ( ( de = mxGetField(marray,0,"dd") ) != NULL ) mxDestroyArray(de);
  de  = mxCreateDoubleMatrix(1,ii.nn,mxREAL);
  dep = mxGetPr(de);
  for ( kk = 0 ; kk < ii.nn ; kk ++ ) {
    * ( dep ++ ) = ii.dd[kk];
  }
  mxSetField(marray,0,"dd", de);
}

int index_length(Index ii) {
  return ii.nn;
}

bool index_isempty(Index ii) {
  return ( ii.nn == 0 );
}

bool index_subindex(Index ii, Index jj) {
  int *idp, *idend, id;
  int *jdp, *jdend, jd;

  idp = ii.dd; idend = ii.dd + ii.nn;   
  jdp = jj.dd; jdend = jj.dd + jj.nn; 

  while (1) {
    if ( idp == idend ) {
      return 1;
    } else if ( jdp == jdend ) {
      return 0;
    }
    id = * idp;
    jd = * jdp;
    if ( id == jd ) {
      idp ++; jdp ++;
    } else if ( id < jd ) {
      return 0;
    } else {
      jdp ++; 
    }
  }
}

Index index_union(Index ii, Index jj) {
  Index oo;
  int *idp, *idend, id;
  int *jdp, *jdend, jd;
  int *odp;

  oo = index_new(ii.nn + jj.nn);

  idp = ii.dd; idend = ii.dd + ii.nn;   
  jdp = jj.dd; jdend = jj.dd + jj.nn; 
  odp = oo.dd; 

  while (1) {
    if ( idp == idend ) {
      while ( jdp < jdend ) {
        * ( odp ++ ) = * ( jdp ++ );
      }
      break;
    } else if ( jdp == jdend ) {
      while ( idp < idend ) {
        * ( odp ++ ) = * ( idp ++ );
      }
      break;
    }
    id = * idp;
    jd = * jdp;
    if ( id == jd ) {
      * ( odp ++ ) = id;
      idp ++; 
      jdp ++; 
    } else if ( id < jd ) {
      * ( odp ++ ) = id;
      idp ++; 
    } else {
      * ( odp ++ ) = jd;
      jdp ++; 
    }
  }
  oo.nn  = (int) ( odp - oo.dd );
  oo.dd  = mxRealloc(oo.dd, sizeof(int)*oo.nn);
  return oo;
}

Index index_intersect(Index ii, Index jj) {
  Index oo;
  int *idp, *idend, id;
  int *jdp, *jdend, jd;
  int *odp; 

  oo = index_new(ii.nn + jj.nn);

  idp = ii.dd; idend = ii.dd + ii.nn;   
  jdp = jj.dd; jdend = jj.dd + jj.nn; 
  odp = oo.dd; 

  while (1) {
    if ( idp == idend ) {
      /*
      while ( jdp < jdend ) {
        * ( odp ++ ) = * ( jdp ++ );
      }
      */
      break;
    } else if ( jdp == jdend ) {
      /*
      while ( idp < idend ) {
        * ( odp ++ ) = * ( idp ++ );
      }
      */
      break;
    }
    id = * idp;
    jd = * jdp;
    if ( id == jd ) {
      * ( odp ++ ) = id;
      idp ++; 
      jdp ++; 
    } else if ( id < jd ) {
      /*
      * ( odp ++ ) = id;
      */
      idp ++; 
    } else {
      /*
      * ( odp ++ ) = jd;
      */
      jdp ++; 
    }
  }
  oo.nn  = (int) ( odp - oo.dd );
  oo.dd  = mxRealloc(oo.dd, sizeof(int)*oo.nn);
  return oo;
}

Index index_diff(Index ii, Index jj) {
  Index oo;
  int *idp, *idend, id;
  int *jdp, *jdend, jd;
  int *odp; 

  oo = index_new(ii.nn + jj.nn);

  idp = ii.dd; idend = ii.dd + ii.nn;   
  jdp = jj.dd; jdend = jj.dd + jj.nn; 
  odp = oo.dd; 

  while (1) {
    if ( idp == idend ) {
      /*
      while ( jdp < jdend ) {
        * ( odp ++ ) = * ( jdp ++ );
      }
      */
      break;
    } else if ( jdp == jdend ) {
      while ( idp < idend ) {
        * ( odp ++ ) = * ( idp ++ );
      }
      break;
    }
    id = * idp;
    jd = * jdp;
    if ( id == jd ) {
      /*
       * ( odp ++ ) = id;
      */
      idp ++; 
      jdp ++; 
    } else if ( id < jd ) {
      * ( odp ++ ) = id;
      idp ++; 
    } else {
      /*
      * ( odp ++ ) = jd;
      */
      jdp ++; 
    }
  }
  oo.nn  = (int) ( odp - oo.dd );
  oo.dd  = mxRealloc(oo.dd, sizeof(int)*oo.nn);
  return oo;
}

Index index_xor(Index ii, Index jj) {
  Index oo;
  int *idp, *idend, id;
  int *jdp, *jdend, jd;
  int *odp;

  oo = index_new(ii.nn + jj.nn);

  idp = ii.dd; idend = ii.dd + ii.nn;   
  jdp = jj.dd; jdend = jj.dd + jj.nn; 
  odp = oo.dd; 

  while (1) {
    if ( idp == idend ) {
      while ( jdp < jdend ) {
        * ( odp ++ ) = * ( jdp ++ );
      }
      break;
    } else if ( jdp == jdend ) {
      while ( idp < idend ) {
        * ( odp ++ ) = * ( idp ++ );
      }
      break;
    }
    id = * idp;
    jd = * jdp;
    if ( id == jd ) {
      /* 
      * ( odp ++ ) = id;
      */
      idp ++; 
      jdp ++; 
    } else if ( id < jd ) {
      * ( odp ++ ) = id;
      idp ++; 
    } else {
      * ( odp ++ ) = jd;
      jdp ++; 
    }
  }
  oo.nn  = (int) ( odp - oo.dd );
  oo.dd  = mxRealloc(oo.dd, sizeof(int)*oo.nn);
  return oo;
}

